import fnmatch
import hashlib
import json
import os
import re
from http import HTTPStatus

from fovus.constants.cli_action_runner_constants import USER_CONFIG_EMAIL_UPDATE_MESSAGE
from fovus.constants.cli_constants import DEFAULT_USER_CONFIG_FILE_PATH, EMAIL, USERNAME
from fovus.constants.util_constants import UTF8
from fovus.exception.user_exception import UserException
from fovus.util.util import Util

FOVUS_JOB_INFO_FOLDER = ".fovus"
VALID_TASK_FOLDER_REGEX = re.compile("^[a-zA-Z0-9][a-zA-Z0-9_-]*$")


class FileUtil:
    @staticmethod
    def get_files_in_directory(root_directory_path, include_input_list, exclude_input_list):
        task_directories = FileUtil._get_task_directories(root_directory_path, include_input_list, exclude_input_list)
        FileUtil._validate_task_directory_names(task_directories)
        local_filepath_list = []
        total_file_size_bytes = 0
        for directory_path, _, files in os.walk(root_directory_path):
            for file in files:
                local_filepath = os.path.join(directory_path, file)
                if FileUtil.should_ignore_path(file) or not FileUtil.include_exclude_allows_path(
                    os.path.relpath(local_filepath, root_directory_path), include_input_list, exclude_input_list
                ):
                    continue
                local_filepath_list.append(local_filepath)
                total_file_size_bytes += os.path.getsize(local_filepath)
        return local_filepath_list, len(task_directories), total_file_size_bytes

    @staticmethod
    def _get_task_directories(root_directory_path, include_input_list, exclude_input_list):
        directories = []
        for entry in os.scandir(root_directory_path):
            if (
                entry.is_dir()
                and entry.name != FOVUS_JOB_INFO_FOLDER
                and FileUtil.include_exclude_allows_path(entry.name + os.sep, include_input_list, exclude_input_list)
            ):
                directories.append(entry.name)
        return directories

    @staticmethod
    def _validate_task_directory_names(task_directory_names):
        invalid_dirs = [
            task_directory
            for task_directory in task_directory_names
            if not re.match(VALID_TASK_FOLDER_REGEX, task_directory)
        ]
        if invalid_dirs:
            raise UserException(
                HTTPStatus.BAD_REQUEST,
                FileUtil.__name__,
                "Task folders must only contain the following characters (commas not included): a-Z, 0-9, -, _. "
                "In addition, task folders must not start with a hyphen - or underscore _. "
                "The following task folder names are invalid: " + str(invalid_dirs),
            )

    @staticmethod
    def should_ignore_path(file_name):
        if FileUtil._path_contains_directory(file_name, FOVUS_JOB_INFO_FOLDER):
            return True
        if Util.is_unix():
            return FileUtil._is_swap_or_hidden_file_unix(file_name)
        if Util.is_windows():
            return FileUtil._is_temp_or_hidden_file_windows(file_name)
        raise UserException(
            HTTPStatus.BAD_REQUEST,
            FileUtil.__name__,
            "Unsupported operating system. Only Windows and Unix are supported.",
        )

    @staticmethod
    def include_exclude_allows_path(file_relative_path, include_list, exclude_list):
        if include_list and not FileUtil._should_include_file(file_relative_path, include_list):
            return False
        if exclude_list and FileUtil._should_exclude_file(file_relative_path, exclude_list):
            return False
        return True

    @staticmethod
    def _should_include_file(file_name, include_input_list):
        for include_input in include_input_list:
            if fnmatch.fnmatch(file_name, include_input):
                return True
        return False

    @staticmethod
    def _should_exclude_file(file_name, exclude_input_list):
        for exclude_input in exclude_input_list:
            if fnmatch.fnmatch(file_name, exclude_input):
                return True
        return False

    @staticmethod
    def _path_contains_directory(filepath, directory):
        return directory in filepath.split(os.path.sep)

    @staticmethod
    def directory_contains_directory(directory_path, directory):
        return directory in os.listdir(directory_path)

    @staticmethod
    def _is_swap_or_hidden_file_unix(filepath):
        is_hidden = os.path.basename(filepath).startswith(".")
        is_swap = os.path.basename(filepath).endswith("~")
        return is_hidden or is_swap

    @staticmethod
    def _is_temp_or_hidden_file_windows(filepath):  # pylint: disable=unused-argument
        return False  # TODO

    @staticmethod
    def open(filepath, mode="r", encoding=UTF8):
        if mode.startswith("r") and not FileUtil.file_exists(filepath):
            raise UserException(
                HTTPStatus.BAD_REQUEST,
                FileUtil.__name__,
                f"File does not exist: {filepath}",
            )
        if "b" in mode:
            return open(filepath, mode)
        return open(filepath, mode, encoding=encoding)

    @staticmethod
    def file_exists(filepath):
        return os.path.exists(filepath)

    @staticmethod
    def get_file_s3_hash(filepath, chunk_size=5 * 1024 * 1024) -> str:
        md5_bytes_list: list[bytes] = []

        with FileUtil.open(filepath, "rb") as file:
            while True:
                data = file.read(chunk_size)
                if not data:
                    break
                md5_bytes_list.append(hashlib.md5(data, usedforsecurity=False).digest())

        if len(md5_bytes_list) < 1:
            raise UserException(
                HTTPStatus.BAD_REQUEST,
                FileUtil.__name__,
                f"md5 could not be read for: {filepath}",
            )

        if len(md5_bytes_list) == 1:
            return md5_bytes_list[0].hex()

        concatenated_md5_bytes = b"".join(md5_bytes_list)
        s3_md5_string = hashlib.md5(concatenated_md5_bytes, usedforsecurity=False).hexdigest()
        return "-".join([s3_md5_string, str(len(md5_bytes_list))])

    @staticmethod
    def get_file_hash(filepath):
        with FileUtil.open(filepath, "rb") as file:
            return hashlib.md5(file.read(), usedforsecurity=False).hexdigest()

    @staticmethod
    def create_missing_directories(local_root_directory_path, directories):
        for directory in directories:
            local_directory_path = os.path.join(local_root_directory_path, directory)
            os.makedirs(local_directory_path, exist_ok=True)

    @staticmethod
    def windows_to_unix_path(path):
        return path.replace("\\", "/")

    @staticmethod
    def get_default_user_config():
        default_user_config = {}
        with FileUtil.open(DEFAULT_USER_CONFIG_FILE_PATH) as file:
            default_user_config = json.load(file)
        if USERNAME in default_user_config:
            print(
                f"{USER_CONFIG_EMAIL_UPDATE_MESSAGE}"
                "\nSince the user config being used is the default config, "
                f"automatically updating the {USERNAME} key to the {EMAIL} key. "
                f"\n\t- If both {EMAIL} and {USERNAME} are present, the {EMAIL} key/value pair will be kept and "
                f"the {USERNAME} key/value pair will be removed."
            )
            Util.keep_prioritized_key_value_in_dict(default_user_config, prioritized_key=EMAIL, fallback_key=USERNAME)
            with FileUtil.open(DEFAULT_USER_CONFIG_FILE_PATH, "w") as file:
                json.dump(default_user_config, file, indent=4, sort_keys=True)
        return default_user_config
